/******************************************************************************
 * FILE:
 *    crt0.S
 *
 * UC ARCHITECTURE:
 *    PowerPC
 *
 * UC FAMILY:
 *    SPC58
 *
 * DESCRIPTION:
 *    'C' runtime environment startup file for SPC58 PowerPC architecture.
 *
 *    Its main duty
 *       1. To initialize Stack area needed for 'C' code
 *       2. To clear BSS (not-initialized) data present in the code
 *       3. To initialize DATA (initialized) data present in the code
 *       4. To initialize ECC in RAM memories on SPC58 microcontrollers
 *       5. To initialize Core registers according to EABI specification
 *
 *    Core Exception support
 *       1. Initialization of Cores' Exception system
 *       2. Implementation of External Interrupt handler using SW mode
 *          - SW Handler is shared in between the Cores
 *          - SW handler is multicore capable (reads current CoreId in the
 *            runtime and acts accordingly in its Prolog and Epilog code)
 *
 *    Crt0 startup on multicore systems is driven by a configuration provided
 *    by a corresponding linker file. The configuration contains following items
 *       1. STACK base address (where STACK starts) and its size
 *       2. Clear and copy table addresses (fixed format expected)
 *       3. SDA, SDA2 EABI register (small data/const data support)
 *    Configuration table contains one entry for each Core present in uC
 *
 *    Crt0 provides two startup hooks the user can use to implement early HW
 *    initialization.
 *       1. Crt0PreInit() hook called:
 *          - after STACK area on given core is initialized on the current Core
 *          - before Clear and Copy table operation is executed
 *            ! HW initialization code cannot rely on global data (BSS or DATA)
 *       2. Crt0PostInit() hook called
 *          - when all startup initialization is finished (data can be used)
 *          - before calling _main() application entry point
 *
 * NOTES:
 *    c++ support: not supported
 *
 * FILE HISTORY:
 *    V1.0    23.09.2017     RO
 *       > Initial version
 *    V2.0    25.01.2018     RO
 *       > _main entry point renamed to shared_main to better
 *          reflect shared property of the entry function among cores
 *       > Word copy routine for init data added in addition to
 *          Byte copy routine. Selection which one to use is decided
 *          based on defined criteria of SRC & DST address alignment and
 *          SIZE word multiples before each entry copy operation.
 *    V3.0    09.02.2018     RO
 *       > BMHD section moved to "uc" a dedicated file in "uc" folder.
 *
 ******************************************************************************
 * Copyright (C) 2015-2018 HighTec EDV-Systeme GmbH.  All rights reserved.
 * This is proprietary software. Any use of the software requires a written
 * license agreement with HighTec EDV-Systeme GmbH. Please contact for
 * detailed license information: info@hightec-rt.com.
 *****************************************************************************/


/* ============================================================================
 * SYMBOLS & DEFINES & MACROS
 * ==========================================================================*/

/* Configuration Table offsets
 * ! must correspond to __crt0_config structure defined in the Linker file */
#define STACK                0x00
#define STACK_SIZE           0x04
#define CLEAR_TABLE          0x08
#define COPY_TABLE           0x0C
#define SDA                  0x10
#define SDA2                 0x14
#define CRT0CONFIG_SIZE      0x18           /* size of the init structure */

/* CLEAR table structure offsets
 * ! offsets must correspond to the Linker __clear_table structure */
#define CLEAR_TABLE_DST      0x00           /* start of the area to clear */
#define CLEAR_TABLE_SIZE     0x04           /* size of the area in bytes */

/* COPY table structure offsets
 * ! offsets must correspond to the Linker __copy_table structure */
#define COPY_TABLE_SRC       0x00           /* source address */
#define COPY_TABLE_DST       0x04           /* destination of the copy */
#define COPY_TABLE_SIZE      0x08           /* how many bytes to copy */

/* external symbols required by CRT0 */
.extern shared_main, STT_FUNC, 0            /* shared main entry to application */
.extern __crt0_config, STT_OBJECT, 0        /* crt0 config structure from linker */
.extern IVOR00, STT_OBJECT, 0               /* Core Exception table base from uC */
.extern ls_regs_init, STT_FUNC, 0         /* lock-step core initialization from uC */

/* Core MSR register
 * - enable Machine check and Critical check exceptions bits set */
#define MSR                  0x00021002

/* Exported symbols from CRT0
 * _crt0_reset : entry point used to configure entry point for inactive Cores
 *               that will be started from the Reset Core */
.global _crt0_reset
.type _crt0_reset STT_FUNC

/* lwi: Immediate register load to avoid repetition of two instruction
 *      when 32bit value load needs to be loaded to GPR register */
lwi: .macro register, value
    e_lis  \register, \value@h
    e_or2i \register, \value@l
.endm


/* ============================================================================
 * CRT0 CODE
 * Here execution starts after the Reset. Code is encapsulated in a dedicated
 * user section .crt0
 * ==========================================================================*/

/* crt0 shared code section */
	.section .crt0, "avx"

/* crt0 reset entry point (shared code)
 * this reset entry is used to initialize reset vectors in inactive cores
 * within BSP module */
_crt0_reset:

/* Clear general purpose registers VLE accessible (asm loop directive)
 * using gcc macro to make it shorter for registers accessible in vle mode */
.irp gpr,r0,r1,r2,r3,r4,r5,r6,r7,r24,r25,r26,r27,r28,r29,r30,r31
	se_li \gpr, 0
.endr

/* Clear general purpose registers BOOKE accessible (asm loop directive)
 * using gcc macro to make it shorter for registers not accessible in vle mode */
.irp gpr,r9,r10,r11,r12,r13,r14,r15,r16,r17,r18,r19,r20,21,r22,r23
	e_li \gpr, 0
.endr

/* ----------------------------------------------------------------------------
 * CORE LOCK-STEP REGISTERS INIT
 * uC derivative specific
 * --------------------------------------------------------------------------*/

	e_bl      ls_regs_init

/* ----------------------------------------------------------------------------
 * CORE REGISTER INIT
 * initialization of important Core registers at the very beginning
 * --------------------------------------------------------------------------*/

/* Configure MSR register */
	lwi       r3, MSR
	mtmsr     r3

/* Initialize IVOR register for exception handling in the beginning
 * of the startup in case something goes wrong */
	lwi       r3, IVOR00
    mtspr     63, r3;

/* Branch prediction enable (BUCSR register) */
	e_li      r3, 0x201
	mtspr     1013, r3

/* ----------------------------------------------------------------------------
 * CRT0 CONFIG TABLE READ
 * Config table contains parameters steering crt0 operations.
 * It is prepared by the linker file with the knowledge of the final placement
 * --------------------------------------------------------------------------*/

/* Get CRT0 Config entry base for current Core[x] */
	lwi       r24, __crt0_config            /* base of the config structure */
	mfspr     r25, 286                      /* read CoreID SFR register */
	mulli     r25, r25, CRT0CONFIG_SIZE     /* get the right base (offset) */
	add       r24, r24, r25                 /* adjust the entry pointer */

/* ----------------------------------------------------------------------------
 * STACK INITIALIZATION
 * Mandatory operation before calling 'C' Pre-Init code
 * Two things
 *   1. correct ECC init for complete Stack area
 *   2. Stack pointer init used by 'C' code
 * --------------------------------------------------------------------------*/

/* STACK area initialization (ECC init by at least 64bit Write!) */
	e_lwz     r6, STACK_SIZE(r24)           /* stack size in Bytes */
	e_lwz     r5, STACK(r24)                /* STACK base pointer to r3 */
	se_sub    r5, r6                        /* get STACK bottom as base for clear routine */
	e_bl      clear64bit_exec               /* call clear routine */

/* Stack pointer init before jump to 'C' pre-init code */
	e_lwz     r1, STACK(r24)                /* Stack pointer */
	e_stwu    r0, -8(r1)                    /* adjust Stack according to EABI stack frame */

/* ----------------------------------------------------------------------------
 * CRT0 PRE-INIT 'C' USER CODE
 * Chance for user to execute HW init at the very beginning, before
 * longer operations take place, like clear of memories and copy init data.
 * Current CoreId is passed as parameter
 * (!) pre-init code must rely only on Stack variables
 * --------------------------------------------------------------------------*/

	mfspr     r3, 286
	e_bl      Crt0PreInit

/* ----------------------------------------------------------------------------
 * CLEAR .BSS SECTIONS
 * Clear of data in 8B chunks due to ECC int requirements (64bit Write).
 * Regions to clear are passed by structure within the __crt0_config
 * --------------------------------------------------------------------------*/

/* Clear registers used as source in 64bit Clear operation */
	xor       r30, r30, r30
	xor       r31, r31, r31

/* load Clear table base address from the __crt0_config
 * and check a NULL pointer to skip this step if found
 * (!) do it here in case crt0_config has NULL in place of pointer to
 *     CLEAR_TABLE, then the SIZE read in next step would be wrong */
	e_lwz     r7, CLEAR_TABLE(r24)
	se_bmaski r0, 0                         /* prepare -1 value */
	se_cmp    r7, r0                        /* compare values */
	beq       clear_bss_end                 /* end table delimiter found, jump to the end */

/* process table items untill end delimiter is found */
clear_bss_item_process:
	e_lwz     r6, CLEAR_TABLE_SIZE(r7)      /* get item size */
	se_bmaski r0, 0                         /* prepare -1 value (can't be omitted) */
	se_cmp    r6, r0                        /* compare values */
	beq       clear_bss_end                 /* end table delimiter found, jump to the end */

/* item can be equal to '0', in such case move to next item */
	xor       r0, r0, r0                    /* clear r0 */
	se_cmp    r6, r0                        /* if size of the item == 0 */
	beq       clear_bss_item_next           /* go to next item*/

/* prepare destination pointer and call a clear routine */
	e_lwz     r5, CLEAR_TABLE_DST(r7)       /* get (*dst) pointer */
	e_bl      clear64bit_exec               /* call clear routine */

/* move pointer to next item in the table */
clear_bss_item_next:
	se_addi   r7, 0x8                       /* move to next item */
	se_b      clear_bss_item_process        /* go to process next item */

/* end of BSS clear operation */
clear_bss_end:

/* ----------------------------------------------------------------------------
 * CLEAR .DATA SECTIONS
 * Clear memory allocated by initialized data before Byte copy from Flash
 * to RAM is executed. It is a mandatory operation to initialize ECC system,
 * otherwise byte copy would fail on byte write to RAM due to ECC error.
 * --------------------------------------------------------------------------*/

/* load Copy table base address from __crt0_config
 * and check against NULL pointer to skip this step if found
 * NOTE: Jump over copy routine too if NULL pointer found */
	e_lwz     r7, COPY_TABLE(r24)
	se_bmaski r0, 0                         /* prepare -1 value */
	se_cmp    r7, r0                        /* compare values */
	beq       copy_data_end                 /* end table delimiter found, jump to the end */

/* process table items untill end delimiter is found */
clear_data_item_process:
	e_lwz     r6, COPY_TABLE_SIZE(r7)       /* Get item size */
	se_bmaski r0, 0                         /* prepare -1 value (can't be omitted */
	se_cmp    r6, r0                        /* compare values */
	beq       clear_data_end                /* end delimiter found, jump to end */

/* check item size, move to next item if equal to '0' */
	xor       r0, r0, r0                    /* clear r0 */
	se_cmp    r6, r0                        /* if size of the item == 0 */
	beq       clear_data_item_next          /* go to next item*/

/* prepare destination pointer and call clear routine */
	e_lwz     r5, COPY_TABLE_DST(r7)        /* get (*dst) pointer */
	e_bl      clear64bit_exec               /* call clear routine */

/* move pointer to next item in the table */
clear_data_item_next:
	se_addi   r7, 0xC                       /* move to next item */
	se_b      clear_data_item_process       /* go to process next item */

clear_data_end:

/* ----------------------------------------------------------------------------
 * COPY INITIALIZED DATA
 * Initialization of init sections provided in Copy table.
 * The Copy routine selects between two Copy variants, Byte copy or Word copy.
 * Decision which one to use depends solely on SRC, DST and SIZE values read
 * from copy table entry.
 * 1. If SRC, DST and SIZE are Word compliant (word remainder = 0) => Word copy
 * 2. Otherwise => Byte copy
 * --------------------------------------------------------------------------*/

/* load Copy Table base address from __crt0_config */
	e_lwz     r7, COPY_TABLE(r24)

/* process table items until end delimiter is found */
copy_data_item_process:
	e_lwz     r6, COPY_TABLE_SIZE(r7)       /* Get item size */
	se_bmaski r0, 0                         /* prepare -1 value */
	se_cmp    r6, r0                        /* compare values */
	beq       copy_data_end                 /* end delimiter found, jump to end */

/* check item size, move to next item if equal to '0' */
	xor       r0, r0, r0                    /* clear r0 */
	se_cmp    r6, r0                        /* if size of the item == 0 */
	beq       copy_data_item_next           /* go to next item*/

/* prepare Source and Destination pointers, size is already in r6 */
	e_lwz     r4, COPY_TABLE_SRC(r7)        /* get (*src) pointer */
	e_lwz     r5, COPY_TABLE_DST(r7)        /* get (*dst) pointer */

/* prepare Link register to go to next table item after copy routine */
	lwi		  r0, copy_data_item_next       /* take address of copy_data_item_next */
	mtspr	  8, r0                         /* move it to Link Register - address
	                                           of the return from copy sub routine */

/* Select right copy routine, Byte or Word copy, based on following criteria
 * (SRC word align) & (DST word aligned) & (size in word multiples) => Word copy
 * otherwise Byte copy */
 	e_ori     r0, r4, 0                     /* r0 <= (r4 | 0) */
 	se_or     r0, r5                        /* r0 <= (r0 | r5) */
 	se_or     r0, r6                        /* r0 <= (r0 | r6) */
 	e_rlwinm  r0,r0,0,30,31                 /* Word remainder of the (r4 | r5 | r6) */
 	se_cmpi   r0, 0                         /* one of the items not Word compatible */
 	bne       byte_copy_exec                /* do Byte copy OR */
	se_b      word_copy_exec                /* do Word copy */

/* move pointer to next item in the table */
copy_data_item_next:
	se_addi   r7, 0xC                       /* move to next item */
	se_b      copy_data_item_process        /* go to process next item */

copy_data_end:

/* ----------------------------------------------------------------------------
 * EABI REGISTERS INIT
 * Initialize other GPRS register required EABI apart from stak that is
 * already initialized.
 * --------------------------------------------------------------------------*/

/* Small data pointers initialization */
	e_lwz     r13, SDA(r24)                 /* SDATA pointer ('rw' data) */
	e_lwz     r2,  SDA2(r24)                /* SDATA2 pointer ('ro' data) */

/* ----------------------------------------------------------------------------
 * CRT0 POST-INIT 'C' USER CODE
 * Chance for user to execute specific code before jump to shared main
 * Current CoreId is passed as parameter
 * --------------------------------------------------------------------------*/
	mfspr     r3, 286
	e_bl      Crt0PostInit

/* ----------------------------------------------------------------------------
 * CRT0 END
 * Jump to main() entry point shared to all cores
 * Pass CoreID as parameter
 * --------------------------------------------------------------------------*/

_crt0_end:
	mfspr     r3, 286                       /* read CoreID SFR register */
	e_bl      shared_main                   /* jump to shared_main */

/* Infinitive loop, in case of return
 * post run phase might be postioned here instead of infinitive loop */
	se_b      .


/* ============================================================================
 * WEAK FUNCTIONS
 * Empty replacement of pre-init and post-init functions called from crt0,
 * if the user does not implement them. They do nothing, just return back
 * to caller.
 * ==========================================================================*/

/* CRT0 Pre-Init function */
	.weak Crt0PreInit
	.type Crt0PreInit STT_FUNC
Crt0PreInit:
	se_blr

/* CRT0 Post-Init function */
	.weak Crt0PostInit
	.type Crt0PostInit STT_FUNC
Crt0PostInit:
	se_blr


/* ============================================================================
 * LOCAL ROUTINES
 * ==========================================================================*/

/* ----------------------------------------------------------------------------
 * FUNCTION: clear64bit_exec
 *    Implements 64bit write to destination memory by using e_stmw instruction.
 *    This instruction uses r30, r31 gprs registers in 64bit write. They shall
 *    be cleared/set by user before calling this routine
 *    Function does write pointer and size alignment to 64bit that is required
 *    by the e_stmw.
 *    Example:
 *       1. clear 2B at address 0x04 -> write to range 0x00-0x07 (1x64bit)
 *       2. clear 6B at address 0x06 -> write to range 0x00-0x0F (2x64bit)
 * INPUT:
 *    r5 : destination pointer
 *    r6 : size in bytes
 * --------------------------------------------------------------------------*/

clear64bit_exec:
	e_rlwinm  r0,r5,0,29,31                 /* remainder of (*dst) DW alignment */
	se_add    r6, r0                        /* add remainder to item (size) */
	se_addi   r6, 7                         /* add 7 bytes to (size) */
	se_srwi   r6, 3                         /* rotate right to get (size) in DW */
	e_rlwinm  r5,r5,0,0,28                  /* (*dst) DW alignment */
	mtctr     r6                            /* move (size) in DW to loop counter */
clear64bit_loop:
	e_stmw    r30, 0(r5)                    /* 64bit write to (*dst) */
	e_addi    r5,r5,0x8                     /* move (*dst) pointer by 8B (DW) */
	bdnz 	  clear64bit_loop               /* decrement counter,loop if not 0 */
	se_blr


/* ----------------------------------------------------------------------------
 * FUNCTION: byte_copy_exec
 *    It copies data from source to destination Byte by Byte.
 * INPUT:
 *    r4 : source pointer
 *    r5 : destination pointer
 *    r6 : size in bytes
 * --------------------------------------------------------------------------*/

byte_copy_exec:
	mtctr     r6                            /* move (size) in Bytes to loop counter */
    e_addi    r5, r5, -1                    /* adjust for stbu instruction use */
    e_addi    r4, r4, -1                    /* adjust for lbzu instruction use */
byte_copy_loop:
	e_lbzu    r3, 1(r4)                     /* load byte, update pointer after */
	e_stbu    r3, 1(r5)                     /* store byte, update pointer after */
	bdnz      byte_copy_loop                /* decrement counter,loop if not 0 */
	se_blr                                  /* return to caller */


/* ----------------------------------------------------------------------------
 * FUNCTION: word_copy_exec
 *    It copies data from source to destination in 4B chunks (words). Size
 *    value needs to be Word adjusted as it comes in Bytes.
 * INPUT:
 *    r4 : source pointer
 *    r5 : destination pointer
 *    r6 : size in bytes
 * --------------------------------------------------------------------------*/

word_copy_exec:
	se_srwi   r6, 2                         /* rotate right to get (size) in Word */
	mtctr     r6                            /* move (size) in Words to loop counter */
    e_addi    r5, r5, -4                    /* adjust for stwu instruction use */
    e_addi    r4, r4, -4                    /* adjust for lbwu instruction use */
word_copy_loop:
	e_lwzu    r3, 4(r4)                     /* load word, update pointer after */
	e_stwu    r3, 4(r5)                     /* store word, update pointer after */
	bdnz      word_copy_loop                /* decrement counter,loop if not 0 */
	se_blr                                  /* return to address in link register */
